#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 19 17:11:30 2020

@author: 
"""

import tensorflow as tf
import tensorflow.keras as keras
from classifier import Classifier
#from cnn import CifarInput, ImageNetInput, Output

from module import Conv2D, Dense, BatchNormalization

import numpy as np
import math
from collections import defaultdict


@tf.custom_gradient
def rgf_relu(x):
    o1 = tf.nn.relu(x)
    o2 = tf.keras.activations.elu(x, alpha=0.1)
    
    def grad(dy):
        return tf.gradients(o2, [x], grad_ys=[dy])

    return o1, grad     


def inverse_sigmoid(x):
    
    y = np.log(x) - np.log(1.0-x)
    
    return y


def expand_sum(a, b):
    a = np.asarray(a)
    b = np.asarray(b)
    temp = np.zeros_like(b)
    temp[:a.size] = a
    c = temp + b
    return c


def featuremap_size(input_shape):
    input_shape = input_shape if isinstance(input_shape, (tuple, list)) else [input_shape]    
    channels = [shape[-1] for shape in input_shape]

    height = input_shape[0][1]
    width = input_shape[0][2]
    return channels, height, width


class EdgeInitializer(tf.keras.initializers.Initializer):


    def __init__(self, mean, stddev):
        self.mean = mean
        self.stddev = stddev
        
        
    def __call__(self, shape, dtype=None, **kwargs):
        
        value = tf.random.normal(shape, mean=self.mean, stddev=self.stddev, dtype=dtype)
        mag = tf.reduce_mean(tf.math.abs(value), axis=2, keepdims=True)
        value = value + tf.math.sign(value) * mag
        return value
 

    
class Edge(keras.layers.Layer):
    
    def __init__(self, filters, sparsity_scale=0.000001, l2=0.00005, name='edge'):

        super(Edge, self).__init__(name=name)

        self.filters = filters
        self.sparsity_scale = tf.Variable(sparsity_scale, trainable=False, dtype='float32')        
        self.l2= l2
        self.kernel = None    
        self.total_edges = 0



    def set_sparsity_scale(self, sparsity_scale): 
        
        tf.keras.backend.set_value(self.sparsity_scale, sparsity_scale)

        
    def build(self, input_shape):


        channels, height, width = featuremap_size(input_shape)

        self.channels = channels
        self.height = height
        self.width = width
        
        stddev = math.sqrt(2/self.filters)
        kernel_initializer = EdgeInitializer(mean=0.0, stddev=stddev)
        kernel_regularizer = keras.regularizers.l2(l=self.l2)
        #in_channels : incoming edge per node
        #filters : the number of nodes
        in_channels = sum(channels)
        self.kernel = self.add_weight(shape=(1, 1, in_channels, self.filters), 
                                    initializer=kernel_initializer, 
                                    regularizer=kernel_regularizer,
                                    trainable=True, name=self.name+'_kernel')


        self.total_edges = in_channels * self.filters

        beta_init_value = inverse_sigmoid(0.5/in_channels)
        beta_init = tf.keras.initializers.Constant(beta_init_value)
        self.beta = self.add_weight(shape=(1,1,1, self.filters), initializer=beta_init, 
                                trainable=True, name=self.name+'_beta')

    #active input nodes should have nonzero edges
    def active_input_node(self):
        
        kernel, _, _ = self.get_kernel()
        kernel = tf.squeeze(kernel, axis=[0,1])
        active = tf.where(tf.abs(kernel) > 0, tf.ones_like(kernel), tf.zeros_like(kernel))
        has_output = tf.reduce_sum(active, axis=1)
        
        has_output = tf.split(has_output, num_or_size_splits=self.channels)
        return has_output
    
    def active_output_node(self):
        kernel, _, _ = self.get_kernel()
        kernel = tf.squeeze(kernel, axis=[0,1])
        active = tf.where(tf.abs(kernel) > 0, tf.ones_like(kernel), tf.zeros_like(kernel))
        has_input = tf.reduce_sum(active, axis=0)

        return has_input        
    
    def profile(self):
        
        alpha, beta, reg = self.get_kernel()
        
        nonzeros = tf.math.count_nonzero(alpha, dtype=tf.int32)
        zeros = self.total_edges  - nonzeros
        sparsity_rate = zeros / self.total_edges
        total_edges = self.total_edges
        
        nonzeros = nonzeros.numpy()
        zeros = zeros.numpy()
        sparsity_rate = sparsity_rate.numpy()
        total_edges = total_edges
        
        return zeros, nonzeros, total_edges, sparsity_rate
    

    def flops(self):
        zeros, nonzeros, total_edges, sparsity_rate = self.profile()
        
        spatial_size = self.height * self.width
        flops = int(spatial_size * nonzeros)
        params = int(nonzeros)
        
        return flops, params

        
    def get_kernel(self):
        alpha = self.kernel
        beta = tf.math.sigmoid(self.beta)

        abs_alpha = tf.math.abs(alpha)        
        alpha_l1 = tf.reduce_sum(abs_alpha, axis=2, keepdims=True)

        #alpha_mag = tf.nn.relu(abs_alpha - beta*alpha_sum) 
        alpha_mag = rgf_relu(abs_alpha - beta*alpha_l1) 
        
        alpha = tf.math.sign(alpha) * alpha_mag

        reg = tf.reduce_sum(tf.math.abs(alpha_mag)) #l1-norm
        
        return alpha, beta, reg

            
    def call(self, input_tensor, training=False):

        y = tf.concat(input_tensor, axis=-1) if isinstance(input_tensor, (tuple, list)) else input_tensor
        alpha, beta, reg = self.get_kernel()        
        y = tf.nn.conv2d(y, filters=alpha, strides=1, padding='SAME')        

        scale = self.sparsity_scale 
        self.add_loss(scale * reg)
        self.add_loss(self.l2 * tf.reduce_sum(tf.square(beta)))
        
        return y



class Node(keras.layers.Layer):
    
    def __init__(self, kernel_size=(3,3), strides=(1,1), name='node'):
        
        super(Node, self).__init__(name=name)

        self.kernel_size = kernel_size
        self.strides = strides if isinstance(strides, (tuple, list)) else (strides, strides)
        
        self.batch_norm = BatchNormalization(name=name+'_bn')
        self.activation = tf.nn.relu
        self.conv = None        

    
    def build(self, input_shape):

        channels, height, width = featuremap_size(input_shape)
        in_channels = sum(channels)
        
        self.conv = Conv2D(filters=in_channels, kernel_size=self.kernel_size, 
                              padding='same', strides=self.strides,
                              groups=in_channels, use_bias=False,
                              name=self.name+'_conv')

        self.in_channels = in_channels
        self.height = height
        self.width = width
        
        
    def call(self, input_tensor, training=False):
    
        y = self.batch_norm(input_tensor, training=training)
        y = self.activation(y)
        y = self.conv(y)
        
        return y
    
    
    def flops(self, active_nodes):

        
        kernel_size = self.kernel_size[0] * self.kernel_size[1] 
        spatial_size =  self.height  * self.width 
        stride_size = self.strides[0] *  self.strides[1]
        
        conv_flops = int(active_nodes * kernel_size * spatial_size / stride_size)
        
        conv_params = int(active_nodes * kernel_size)
        
        #batch_norm_flops = int(active_nodes * spatial_size * 2)
        batch_norm_params = int(active_nodes * 2)
        
        flops = int(conv_flops)
        params = int(conv_params + batch_norm_params)
        
        return flops, params        

        
class InternalBlock(keras.layers.Layer):
    
    def __init__(self, filters, kernel_size=(3,3), sparsity_scale=0.000001, name='internal_block'):
        
        super(InternalBlock, self).__init__(name=name)
        
        
        self.kernel_size = kernel_size
        
        # incoming edge operations, filters: the number of nodes within a block            
        self.edge = Edge(filters=filters, sparsity_scale=sparsity_scale, name=name+'_edge')        
        
        # node operations, filters: the number of nodes within a block
        self.node = Node(kernel_size=kernel_size, strides=(1,1), name=name+'_node')


    def call(self, input_tensor, training=False):
        
        y = self.edge(input_tensor, training=training)
        y = self.node(y, training=training)
        
        return y
 
       
    def profile(self):
        return self.edge.profile()
    
    

    def flops(self, active_nodes):

        edge_flops, edge_params = self.edge.flops()
        node_flops, node_params = self.node.flops(active_nodes)
        
        flops = edge_flops + node_flops
        params = edge_params + node_params
        
        return flops, params
    

    
class InputBlock(keras.layers.Layer):
    
    def __init__(self, kernel_size=(3,3), strides=(2,2), name='input_node'):
        
        super(InputBlock, self).__init__(name=name)

        self.kernel_size = kernel_size        
       
        # node operations
        self.node = Node(kernel_size=kernel_size, strides=strides)



    def call(self, input_tensor, training=False):
        
        y = self.node(input_tensor, training=training)
        
        return y

    def flops(self, active_nodes):

        flops, params = self.node.flops(active_nodes)
        
        return flops, params
    


class OutputBlock(keras.layers.Layer):

    def __init__(self, filters, sparsity_scale=0.000001, name='output_node'):
        
        super(OutputBlock, self).__init__(name=name)
        
        self.edge = Edge(filters=filters, sparsity_scale=sparsity_scale, name=name+'_edge')        



    def profile(self):
        return self.edge.profile()
      
        
    def call(self, input_tensor, training=False):
        
        y = self.edge(input_tensor, training=training)        

        return y
    
        
    def flops(self):
        
        flops, params = self.edge.flops()
        has_input = self.edge.active_output_node()
        
        return flops, params, has_input
    

#Graph    
class Stage(keras.layers.Layer):
    
    def __init__(self, blocks, filters, strides=(1,1), alpha=1, sparsity_scale=0.000001, name='stage'):
        
        super(Stage, self).__init__(name=name)
            
        filters = int(filters*alpha)
        self.strides = strides
        
        self.internal_blocks = []
        
        self.input_block = InputBlock(strides=strides, name=name+'_input')
        for i in range(blocks-1):
            block = InternalBlock(filters=filters, sparsity_scale=sparsity_scale, name=name+'_node_%d'%i)
            self.internal_blocks.append(block)
        
        self.output_block = OutputBlock(filters=filters, sparsity_scale=sparsity_scale, name=name+'_output')
        

        
    def call(self, input_tensor, training=False):
    
        y = [self.input_block(input_tensor, training=training)]
        
        for block in self.internal_blocks:
            
            y = y + [block(y, training=training)]
        
        y = self.output_block(y)
        
        return y
    
    
    
    def edge(self):

        edges = [block.edge for block in self.internal_blocks]
        edges.append(self.output_block.edge)

        return edges
    
    
    def profile(self):
        
        zeros_, nonzeros_, total_edges_, sparsity_rate_ = self.output_block.profile()
        zeros = zeros_
        nonzeros = nonzeros_        
        total_edges = total_edges_
        
        for block in self.internal_blocks:
            zeros_, nonzeros_, total_edges_, sparsity_rate_ = block.profile()
            zeros += zeros_
            nonzeros += nonzeros_
            total_edges += total_edges_
        
        sparsity_rate = zeros / total_edges
        
        return zeros, nonzeros, total_edges, sparsity_rate

       
    def flops(self):
        
        flops, params = 0, 0
                
        edges = self.edge()
        has_output = [0] * len(edges)
        for edge in edges:
            has_output_ = edge.active_input_node()
            for i in range(len(has_output_)):
                has_output[i] = has_output[i] + has_output_[i]
            
        active_inputs = None
        blocks = [self.input_block] + self.internal_blocks
        for block, has_output_ in zip(blocks, has_output):
            active_nodes = tf.math.count_nonzero(has_output_, dtype=tf.int32)
            if active_inputs is None: active_inputs = active_nodes 
            
            block_flops, block_params = block.flops(active_nodes=active_nodes)
            flops += block_flops
            params += block_params
        
        output_flops, output_params, has_input = self.output_block.flops()
        active_outputs = tf.math.count_nonzero(has_input, dtype=tf.int32)
        
        flops += output_flops
        params += output_params
        return flops, params, active_inputs, active_outputs

            

    
class Input(keras.layers.Layer):
    
    def __init__(self, filters=32, alpha=1.0, kernel_size=(3,3), strides=(2,2), name='input'):

        super(Input, self).__init__(name=name)
        
        filters = int(filters*alpha)
        self.conv = Conv2D(filters=filters, kernel_size=kernel_size, strides=strides, padding='same')
        
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        
    def build(self, input_shape):
        
        self.height = input_shape[1] 
        self.width = input_shape[2] 
        self.in_channels = input_shape[3] 
        
        
    def call(self, input_tensor, training=False):
        
        y = self.conv(input_tensor)
        
        return y


    def flops(self, active_outputs):      

        kernel_size = self.kernel_size[0] * self.kernel_size[1] 
        spatial_size =  self.height  * self.width 
        stride_size = self.strides[0] * self.strides[1]

        
        flops = int(self.in_channels * active_outputs * kernel_size * spatial_size / stride_size)
        params = int(self.in_channels * active_outputs * kernel_size)

        
        return flops, params
    
    

class Output(keras.layers.Layer):
    
    def __init__(self, units, dropout_rate=1e-3, name='outputs'):
        
        super(Output, self).__init__(name=name)

        self.units = units
        
        self.batch_norm = BatchNormalization(name=name+'_bn')
        self.activation = tf.nn.relu
        self.pooling2d = keras.layers.GlobalAveragePooling2D()
        self.dropout = keras.layers.Dropout(dropout_rate) if dropout_rate > 0 else None
        self.dense = Dense(units)
        

    def build(self, input_shape):
        
        self.height = input_shape[1] 
        self.width = input_shape[2] 


               
    def call(self, input_tensor, training=False):
        
        y = self.batch_norm(input_tensor, training=training)
        y = self.activation(y)
        y = self.pooling2d(y)
        y = self.dropout(y, training=training) if self.dropout else y
        y = self.dense(y)
        
        return y

    def flops(self, active_outputs): # active inputs
        
        batch_norm_params = int(active_outputs * 2)
        
        pooling_flops = self.height * self.width * active_outputs
        dense_flops = active_outputs * self.units
        
        flops = int(pooling_flops + dense_flops)
        params = int(batch_norm_params + active_outputs * self.units + self.units)

        
        return flops, params
    
    
class MobileNet(Classifier):
    
    def __init__(self, classes, input_filters=32, alpha=1.0, sparsity_scale=0.000001, dropout_rate=1e-3):
        
        
        super(MobileNet, self).__init__()
        if classes == 10 or classes == 100: strides=(1,1)
        else: strides = (2,2)                 
        
        self.inputs = Input(filters=input_filters, alpha=alpha, strides=strides)
        
        blocks = (1, 2, 2, 6, 2)
        filters = (64, 128, 256, 512, 1024)
        strides = (1, 2, 2, 2, 2)

        
        self.stages = []
        
        for i, (blocks_, filters_, strides_) in enumerate(zip(blocks, filters, strides)):

            stage = Stage(blocks_, filters_, strides_, alpha=alpha, sparsity_scale=sparsity_scale, name='stage_%d'%i)
            self.stages.append(stage)
            
        self.outputs = Output(units=classes, dropout_rate=dropout_rate)
        

        
    def call(self, input_tensor, training=False):
        
        y = self.inputs(input_tensor, training=training)
        
        for stage in self.stages:
            y = stage(y, training=training)
            
        y = self.outputs(y, training=training)
        
        return y
    

    def profile(self):
        
        zeros = 0; nonzeros = 0; total_edges = 0
        stage_profile = dict()
        
        for stage in self.stages:
            profile_ = stage.profile()
            zeros_, nonzeros_, total_edges_, sparsity_rate_ = profile_  
            
            zeros += zeros_
            nonzeros += nonzeros_
            total_edges += total_edges_
            stage_profile[stage.name] = {'zeros':zeros_, 'nonzeros':nonzeros_, 'total_edges':total_edges_, 'sparsity_rate':sparsity_rate_}
            
            
        sparsity_rate = zeros / total_edges
        
        profile = {'zeros':zeros, 'nonzeros':nonzeros, 'total_edges':total_edges, 'sparsity_rate':sparsity_rate}

        return profile, stage_profile
        
    
    def edge(self):
        edges = []
        for stage in self.stages:
            edges += stage.edge()
        
        return edges
        
    def _build_train_step(self):
        
        self._crossentropy = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True, reduction=tf.keras.losses.Reduction.NONE)
        self.optimizer = tf.keras.optimizers.SGD(learning_rate=0.1, momentum=0.9, nesterov=False)
        

    def flops(self):
        
        stage_flops, stage_params = [], []

        first_active_inputs = None
        for stage in self.stages:
            stage_flops_, stage_params_, active_inputs, active_outputs = stage.flops()
            stage_flops.append(stage_flops_)
            stage_params.append(stage_params_)
           
            if first_active_inputs is None: first_active_inputs = active_inputs
        
        input_flops, input_params = self.inputs.flops(first_active_inputs)        
        print('input', input_flops, input_params)

        for stage, flops_, params_ in zip(self.stages, stage_flops, stage_params):
            print(stage.name, flops_, params_)

        stage_flops = sum(stage_flops)
        stage_params = sum(stage_params)
        
        output_flops, output_params = self.outputs.flops(active_outputs)
        print('output', output_flops, output_params)

        
        flops = input_flops + stage_flops + output_flops
        params = input_params + stage_params + output_params 

        return flops, params