#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 14 15:50:39 2021

@author: 
"""

import numpy as np
import tensorflow as tf
import tensorflow.keras as keras

#from regularizer import L2Regularizer

from module import BatchNormalization
import util

def batch_norm(cfg=None, name='batch_norm'):
    
    if cfg is None:
        return BatchNormalization(name=name)

    bn_type = cfg['batch_norm']
    sparsity_scale = cfg['sparsity_scale']
    rgf = cfg['rgf'] if 'rgf' in cfg.keys() else False
    norm = cfg['norm'] if 'norm' in cfg.keys() else 'l1'
    gamma =  cfg['gamma'] if 'gamma' in cfg.keys() else 0.5
    

    if bn_type == 'differentiable':
        return DifferentiableBatchNormalization(sparsity_scale=sparsity_scale, rgf=rgf, norm=norm, gamma=gamma, name=name)
    else: 
        return BatchNormalization(name=name)


def assure_index_dims(indices):

    indices = tf.squeeze(indices)
    if tf.size(tf.shape(indices)) < 1:
        indices = tf.expand_dims(indices, axis=-1)
    
    return indices

    
class Selection(keras.layers.Layer):
    
    
    def __init__(self, indices=None, name='selection'):
        
        super(Selection, self).__init__(name=name)
        
        self.set_index(indices)

    
    def set_index(self, indices):
        
        if indices is None: 
            self.indices = None
            return
        indices = assure_index_dims(indices)
        
        self.indices = tf.constant(value=indices, name='index')
    
        
    def call(self, input_tensor, training=False):

        if self.indices is None: return input_tensor
        #if tf.size(self.indices) == 0: return None

        y = input_tensor        
        y = tf.concat(y, axis=-1) if isinstance(y, (tuple, list)) else y
        y = tf.gather(y, indices=self.indices, axis=-1)
            
        return y

        
class SparseBatchNormalization(keras.layers.Layer):
    
    def __init__(self, gamma=0.5, momentum=0.9, epsilon=1.0e-5, l2=0.00005, sparsity_scale=0.0001, name='sbn'):
        
        super(SparseBatchNormalization, self).__init__(name=name)
        self.gamma = gamma
        self.momentum=momentum
        self.epsilon=epsilon
        self.l2 = l2
        self.sparsity_scale = tf.Variable(sparsity_scale, trainable=False, dtype='float32')        

        self.alpha = None
        self.beta = None
        self.mask = None
        
        self.batch_norm = None
        self.batch_norm_center = None


    def set_sparsity_scale(self, sparsity_scale): 
        
        tf.keras.backend.set_value(self.sparsity_scale, sparsity_scale)


    def _compute_channels(self, input_shape):

        input_shape = input_shape if isinstance(input_shape, (tuple, list)) else [input_shape]
        channels = [shape[-1] for shape in input_shape]
        self.channels = channels

        return channels


    def build(self, input_shape):

        channels = self._compute_channels(input_shape)
        channels = sum(channels)

        #build scaling factors for batch normalization
        self._build_alpha_beta(channels)

        #build moving average, varaince, center for batch normalization
        self._build_batch_norm(channels)


    def _build_batch_norm(self, channels):
    
        center_init = tf.keras.initializers.Zeros()
        center_reg = keras.regularizers.l2(l=self.l2)
        
        batch_norm_center = self.add_weight(shape=(1,1,1, channels),
                                  initializer=center_init,
                                  regularizer=center_reg,
                                  name='bn_center')
        
        self.batch_norm_center = batch_norm_center
    
        self.batch_norm = BatchNormalization(momentum=self.momentum, 
                                       epsilon=self.epsilon, 
                                       scale=False, center=False)


    def _build_alpha_beta(self):
        raise NotImplementedError()
    

    def get_alpha_beta(self):
        raise NotImplementedError()


    def batch_norm_params(self):
        raise NotImplementedError()
        
    def select_channel(self):
        raise NotImplementedError()

        
    def get_state(self):
        
        alpha, beta, reg = self.get_alpha_beta()
        
        sizes = self.channels
        alphas = tf.split(alpha, num_or_size_splits=sizes, axis=-1)
        nonzeros = [tf.math.count_nonzero(alpha, dtype=tf.int32) for alpha in alphas]
        zeros = [s - n for s, n in zip(sizes, nonzeros)]

        return sizes, nonzeros, zeros, alpha, beta, reg
        
    
    
class DifferentiableBatchNormalization(SparseBatchNormalization):    
    

    def __init__(self, gamma=0.5, momentum=0.9, epsilon=1.0e-5, l2=0.00005, sparsity_scale=0.0001, rgf=False, norm='l1', name='sbn'):
        super(DifferentiableBatchNormalization, self).__init__(gamma=gamma, 
                                                                momentum=momentum, 
                                                                epsilon=epsilon, 
                                                                l2=l2, 
                                                                sparsity_scale=sparsity_scale, 
                                                                name=name)
        self.relu = util.rgf_relu if rgf else tf.nn.relu
        self.norm = norm
        
    def _build_alpha_beta(self, channels):


        alpha_reg = keras.regularizers.l2(l=self.l2)
        
        alpha_init_value = self.gamma *(channels+1.0) / channels
        alpha_init = tf.keras.initializers.Constant(alpha_init_value)
        alpha = self.add_weight(shape=(1, 1, 1, channels), initializer=alpha_init, 
                                    regularizer=alpha_reg,
                                    trainable=True, name='ssr_alpha')
        self.alpha = alpha
         
        beta_init_value = - np.log(channels*channels + channels - 1.0)
        beta_init = tf.keras.initializers.Constant(beta_init_value)
        beta = self.add_weight(shape=(1,), initializer=beta_init, 
                               trainable=True, name='ssr_beta')
        self.beta = beta



    def get_alpha_beta(self):

        alpha = self.alpha
        sig_beta = tf.math.sigmoid(self.beta)

        abs_alpha = tf.math.abs(alpha)        
        alpha_l1 = tf.reduce_sum(abs_alpha)
        alpha_mag = self.relu(abs_alpha - sig_beta*alpha_l1) 
        
        alpha = tf.math.sign(alpha) * alpha_mag

        reg = 0.0
        if self.norm == 'l1':
            reg = tf.reduce_sum(alpha_mag) #l1-norm            

        elif self.norm == 'group':            
            alpha_mags = tf.reshape(alpha, shape=[len(self.channels), -1])
            reg = tf.reduce_sum(util.l2_norm(alpha_mags, axis=-1))
        
        return alpha, sig_beta, reg
    

    def call(self, input_tensor, training=False):
       
        input_tensor = tf.concat(input_tensor, axis=-1) if isinstance(input_tensor, (tuple, list)) else input_tensor
        
        alpha, beta, reg = self.get_alpha_beta()
        
        y = self.batch_norm(input_tensor, training=training)
        y = alpha * y + alpha * self.batch_norm_center

        self.add_loss(self.sparsity_scale * reg)
        self.add_loss(self.l2 * tf.reduce_sum(tf.square(beta)))
        
        return y


    def batch_norm_params(self):
        
        moving_mean = self.batch_norm.moving_mean
        moving_variance = self.batch_norm.moving_variance
        alpha, _, _ = self.get_alpha_beta()
        gamma = alpha
        beta = alpha * self.batch_norm_center
        
        gamma = tf.squeeze(gamma)
        beta = tf.squeeze(beta)
        
        return moving_mean, moving_variance, gamma, beta
    

    def select_channel(self):

        alpha, _, _ = self.get_alpha_beta()
        alpha = tf.squeeze(alpha)
        zero = tf.constant(0, dtype=tf.float32)
        where = tf.not_equal(alpha, zero)
        indices = tf.where(where)
        #indices = tf.squeeze(indices)
        indices = assure_index_dims(indices)
        
        return indices
        