#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 19 17:11:30 2020

@author: 
"""

import numpy as np
import tensorflow as tf
import tensorflow.keras as keras

from classifier import Classifier
from cnn import CifarInput #, ImageNetInput #, Output
from module import Conv2D, Dense

from sparse_batch_norm import Selection, batch_norm, SparseBatchNormalization


        
class BottleNeck(keras.layers.Layer):
    
    def __init__(self, factor=4, growth_rate=12, dropout_rate=0.2, cfg=None, name='bottleneck'):
        
        super(BottleNeck, self).__init__(name=name)        
        filters = factor * growth_rate

        self.batch_norm = batch_norm(cfg, name=name+'bn')
        self.activation = tf.nn.relu
        self.conv = Conv2D(filters=filters, kernel_size=(1,1), strides=(1,1), padding='same')
        self.dropout = keras.layers.Dropout(dropout_rate) if dropout_rate > 0 else None
        
        #for easy visualization and debuging
        self.split = tf.split if cfg else None
        self.num_splits = factor if cfg else None

        
    def call(self, input_tensor, training=False):

        y = self.batch_norm(input_tensor, training=training)
        y = self.activation(y)
        y = self.conv(y)
        y = self.dropout(y, training=training) if self.dropout else y
        y = self.split(y, num_or_size_splits=self.num_splits, axis=-1) if self.split else y
         
        return y
        

    def modules(self, name=None):
        
        modules = [self.batch_norm, self.conv]
        
        return modules

    
class DenseConn(keras.layers.Layer):
    
    def __init__(self, bottle_neck=True, growth_rate=12, dropout_rate=0.2, cfg=None, name='dense_conn'):
    
        super(DenseConn, self).__init__(name=name)
        
        self.selection = Selection()
        self.bottle_neck = BottleNeck(4, growth_rate, dropout_rate, cfg=cfg, name=name+'_bottleneck') if bottle_neck else None
        self.batch_norm = batch_norm(cfg, name=name+'_bn')
        self.activation = tf.nn.relu
        self.conv = Conv2D(filters=growth_rate, kernel_size=(3,3), strides=(1,1), padding='same', name='_conv')
        self.dropout = keras.layers.Dropout(dropout_rate) if dropout_rate > 0 else None


    def call(self, input_tensor, training=False):

        y = self.selection(input_tensor)        
        y = self.bottle_neck(y, training=training) if self.bottle_neck else y
        y = self.batch_norm(y, training=training)
        y = self.activation(y)
        y = self.conv(y)
        y = self.dropout(y, training=training) if self.dropout else y
    
        return y


    def modules(self):

        if self.bottle_neck:
            modules = [self.selection, self.bottle_neck, self.batch_norm, self.conv]
        else:
            modules = [self.selection, self.batch_norm, self.conv]

        return modules

        
#Stage    
class DenseBlock(keras.layers.Layer):
    
    def __init__(self, layers, bottle_neck=True, growth_rate=12, dropout_rate=0.2, cfg=None, name='dense_block'):
        
        super(DenseBlock, self).__init__(name=name)
        
        self.dense_layers = []
        for i in range(layers):
            dense_conn = DenseConn(bottle_neck, growth_rate, dropout_rate, cfg=cfg, name=name+'_dense_conn_%d'%i) 
                
            self.dense_layers.append(dense_conn)
           
            
    def call(self, input_tensor, training=False):
        
        y = input_tensor if isinstance(input_tensor, (tuple, list)) else [input_tensor]
        for layer in self.dense_layers:
            
            y = y + [layer(y, training=training)]

        return y
    
    
    def modules(self):
        modules = self.dense_layers
        return modules
    
    
    
        
class Transition(keras.layers.Layer): 

    def __init__(self, reduction_rate=0.5, dropout_rate=0.2, pool_size=(2,2), strides=(2,2), multiplicity=1, cfg=None, name='trans_block'):

        super(Transition, self).__init__(name=name)
        
        self.reduction_rate = reduction_rate
        self.multiplicity = multiplicity 

        self.selection = Selection()
        self.batch_norm = batch_norm(cfg, name=name+'bn')
        self.activation = tf.nn.relu
        self.conv = None
        self.dropout = keras.layers.Dropout(dropout_rate) if dropout_rate > 0 else None        
        self.pooling2d = keras.layers.AveragePooling2D(pool_size=pool_size, strides=strides, padding='same')

        #for easy visualization and debuging
        self.split = tf.split if cfg else None
        self.num_splits = None
       

    def build(self, input_shape):

        
        input_channels = sum([shape[-1] for shape in input_shape]) if isinstance(input_shape, (tuple, list)) else input_shape[-1]
        #make filters in multiple of 'multiplicity'
        filters = int(np.floor(self.reduction_rate * input_channels/self.multiplicity)*self.multiplicity)
        self.conv = Conv2D(filters=filters, kernel_size=(1,1), strides=(1,1), padding='same')

        self.num_splits = int(filters/self.multiplicity) if self.multiplicity > 1 and self.split else None
        

        
    def call(self, input_tensor, training=False):
        
        y = self.selection(input_tensor)
        y = self.batch_norm(y, training=training)
        y = self.activation(y)
        y = self.conv(y)
        y = self.dropout(y, training=training) if self.dropout else y
        y = self.pooling2d(y)
        y = self.split(y, num_or_size_splits=self.num_splits, axis=-1) if self.split else y
        
        return y
    
    def modules(self):
        modules = [self.selection, self.batch_norm, self.conv]
        return modules
    

class Output(keras.layers.Layer):
    
    def __init__(self, units, cfg=None, name='output'):
        super(Output, self).__init__(name=name)
        
        self.selection = Selection()
        self.batch_norm = batch_norm(cfg, name=name+'bn')
        self.activation = tf.nn.relu
        self.pooling2d = keras.layers.GlobalAveragePooling2D()
        self.dense = Dense(units)
        
    def call(self, input_tensor, training=False):
        
        y = self.selection(input_tensor)
        y = self.batch_norm(y, training=training)
        y = self.activation(y)
        y = self.pooling2d(y)
        y = self.dense(y)
        
        return y
    
    def modules(self):
        modules = [self.selection, self.batch_norm, self.dense]
        return modules
    

class ImageNetInput(keras.layers.Layer):
    
        def __init__(self, filters, kernel_size=(7,7), pool_size=(3,3), strides=(2,2), name='input'):

            super(ImageNetInput, self).__init__(name=name)
            
            self.conv = Conv2D(filters=filters, kernel_size=kernel_size, strides=strides, padding='same', name=name+'_conv')
            self.batch_norm = batch_norm(cfg=None, name=name+'bn')
            self.activation = tf.nn.relu
            self.pooling2d = keras.layers.MaxPooling2D(pool_size=pool_size, strides=strides, padding='same')


        def call(self, input_tensor, training=False):
            
            y = self.conv(input_tensor)
            y = self.batch_norm(y, training=training)
            y = self.activation(y)
            y = self.pooling2d(y)
            
            return y

        def modules(self):
            
            modules = [self.conv, self.batch_norm]
            
            return modules
        
   
class CifarDenseNet(Classifier):
    
    def __init__(self, classes, depth, blocks=3, dropout_rate=0.0, bottle_neck=True, growth_rate=12, reduction_rate=0.5, cfg=None):
        
        super(CifarDenseNet, self).__init__()
        
       
        layers = int((depth-4) / blocks)
        if bottle_neck: layers = int(layers/2)
        
        initial_filters = 2 * growth_rate
        self.input_block = CifarInput(initial_filters)

        self.blocks = []
        for i in range(blocks):
            dense_block = DenseBlock(layers, bottle_neck, growth_rate, dropout_rate, cfg=cfg, name='dense_block_%d'%i)
            self.blocks.append(dense_block)
                       
            if i < blocks-1:
                transition = Transition(reduction_rate, dropout_rate, 
                                        pool_size=(2,2), strides=(2,2), 
                                        multiplicity=growth_rate,
                                        cfg=cfg,
                                        name='trans_block_%d'%i)
                self.blocks.append(transition)
                
        
        outputs = Output(classes, cfg=cfg)
        self.output_block = outputs
        #self.blocks.append(outputs)
   
       
    def call(self, input_tensor, training=False):
        
        y = self.input_block(input_tensor)
        y = tf.split(y, num_or_size_splits=2, axis=-1)
        for block in self.blocks:
            y = block(y, training=training)

        logits = self.output_block(y, training=training)
        
        return logits
    
    
    def modules(self):
        
        modules = [self.input_block] + self.blocks + [self.output_block]
        
        return modules


    def get_batch_norm(self):
        batch_norms = []
        modules = self.modules()
        
        while len(modules) > 0:
            module = modules.pop(0)
            if isinstance(module, SparseBatchNormalization):
                batch_norms.append(module)
            if hasattr(module, 'modules'):
                modules = module.modules() + modules
                   
        return batch_norms


    
class ImageDenseNet(Classifier):

    def __init__(self, classes, blocks, dropout_rate, bottle_neck, growth_rate, reduction_rate, cfg=None):
        
        super(ImageDenseNet, self).__init__()
        
        initial_filters = 2 * growth_rate
        self.inputs = ImageNetInput(initial_filters, kernel_size=(7,7), 
                                    pool_size=(3,3),
                                    strides=(2,2), name='input')


        self.blocks = []
        for i, layers in enumerate(blocks):
            dense_block = DenseBlock(layers, bottle_neck, growth_rate, dropout_rate, cfg=cfg, name='dense_block_%d'%i)
            self.blocks.append(dense_block)
            
            if i < len(blocks)-1:
                transition = Transition(reduction_rate, dropout_rate, 
                                        pool_size=(2,2), strides=(2,2), 
                                        multiplicity=growth_rate, #
                                        cfg=cfg, name='trans_block_%d'%i)
    
                self.blocks.append(transition)

        self.outputs = Output(classes, cfg=cfg)


       
    def call(self, input_tensor, training=False):
        
        y = self.inputs(input_tensor, training=training)
        y = tf.split(y, num_or_size_splits=2, axis=-1)
        for block in self.blocks:
            y = block(y, training=training)

        logits = self.outputs(y, training=training)
        
        return logits
    

    def modules(self):
        
        modules = [self.inputs] + self.blocks + [self.outputs]
        
        return modules


    def get_batch_norm(self):
        batch_norms = []
        modules = self.modules()
        
        while len(modules) > 0:
            module = modules.pop(0)
            if isinstance(module, SparseBatchNormalization):
                batch_norms.append(module)
            if hasattr(module, 'modules'):
                modules = module.modules() + modules
                   
        return batch_norms
       

def build_DenseNet_K12(classes, depth=40, blocks=3, dropout_rate=0.0, cfg=None):
    
    bottle_neck = False
    growth_rate = 12
    reduction_rate = 1
     
    denseNet = CifarDenseNet(classes, 
                             depth=depth, 
                             blocks=blocks, 
                             dropout_rate=dropout_rate, 
                             bottle_neck=bottle_neck, 
                             growth_rate=growth_rate, 
                             reduction_rate=reduction_rate,
                             cfg=cfg)
    
    return denseNet


def build_DenseNet_C_K12(classes, depth=40, blocks=3, dropout_rate=0.0, reduction_rate=1, cfg=None):
    
    bottle_neck = False
    growth_rate = 12
    
     
    denseNet = CifarDenseNet(classes, 
                             depth=depth, 
                             blocks=blocks, 
                             dropout_rate=dropout_rate, 
                             bottle_neck=bottle_neck, 
                             growth_rate=growth_rate, 
                             reduction_rate=reduction_rate,
                             cfg=cfg)
    
    return denseNet

def build_DenseNet_BC_K12(classes, depth=100, blocks=3, dropout_rate=0.0, reduction_rate=0.5, cfg=None):
    
    bottle_neck = True
    growth_rate = 12
    
    denseNet = CifarDenseNet(classes, 
                             depth=depth, 
                             blocks=blocks, 
                             dropout_rate=dropout_rate, 
                             bottle_neck=bottle_neck, 
                             growth_rate=growth_rate, 
                             reduction_rate=reduction_rate,
                             cfg=cfg)
    
    return denseNet


def build_DenseNet_BC_K40(classes, depth=190, blocks=3, dropout_rate=0.0, reduction_rate=0.5):
    
    bottle_neck = True
    growth_rate = 40
    
    denseNet = CifarDenseNet(classes, 
                             depth=depth, 
                             blocks=blocks, 
                             dropout_rate=dropout_rate, 
                             bottle_neck=bottle_neck, 
                             growth_rate=growth_rate, 
                             reduction_rate=reduction_rate)
    
    return denseNet

def build_DenseNet121(classes, dropout_rate=0.0, reduction_rate=0.5, cfg=None):
    
    blocks = (6, 12, 24, 16)
    bottle_neck = True
    growth_rate = 32
    

    denseNet = ImageDenseNet(classes, 
                             blocks=blocks, 
                             dropout_rate=dropout_rate, 
                             bottle_neck=bottle_neck, 
                             growth_rate=growth_rate, 
                             reduction_rate=reduction_rate,
                             cfg=cfg)
    
    return denseNet



def build_DenseNet169(classes, dropout_rate=0.0, reduction_rate=0.5):
    
    blocks = (6, 12, 32, 32)
    bottle_neck = True
    growth_rate = 32
    

    denseNet = ImageDenseNet(classes, 
                             blocks=blocks, 
                             dropout_rate=dropout_rate, 
                             bottle_neck=bottle_neck, 
                             growth_rate=growth_rate, 
                             reduction_rate=reduction_rate)
    
    return denseNet



def build_DenseNet161(classes, dropout_rate=0.0, reduction_rate=0.5):
    
    blocks = (6, 12, 36, 24)
    bottle_neck = True
    growth_rate = 48
    

    denseNet = ImageDenseNet(classes, 
                             blocks=blocks, 
                             dropout_rate=dropout_rate, 
                             bottle_neck=bottle_neck, 
                             growth_rate=growth_rate, 
                             reduction_rate=reduction_rate)
    
    return denseNet


