#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 20 18:31:23 2020

@author: 
"""

import numpy as np
import matplotlib.pyplot as plt

import tensorflow as tf

from tensorflow.keras import datasets


def visualize(original, augmented):
    plt.figure()
    plt.subplot(1,2,1)
    plt.title('Original image')
    plt.imshow(original)
    
    plt.subplot(1,2,2)
    plt.title('Augmented image')
    plt.imshow(augmented)
    


def channel_mean(images):
    mean = []
    num_samples, height, width, num_channels = images.shape
    
    for i in range(num_channels):
        m = np.mean(images[:,:,:,i])
        mean.append(m)

    # cifar-10, for training data: (125.30672, 122.95055, 113.8652)
    # cifiar-100 for training data: [129.30428, 124.07023, 112.43411]
    return mean


def channel_std(images):
    
    std = []
    num_samples, height, width, num_channels = images.shape
    
    for i in range(num_channels):
        s = np.std(images[:,:,:,i])
        std.append(s)

    # cifar-10, for training data: (62.993233, 62.088722, 66.70494)
    # cifiar-100 for training data: [68.17024, 65.391785, 70.4184]
    return std

    
def channel_standardization(images, mean=None, std=None):
    
    images = np.asarray(images, dtype=np.float32)
    
    if mean is None:
        mean = channel_mean(images)
        
    if std is None:
        std = channel_std(images)
    

    num_samples, height, width, num_channels = images.shape
    
    for i in range(num_channels):
        
        images[:,:,:,i] = (images[:,:,:,i] - mean[i]) / std[i]
        
    return images


def augment(image, target):
    
    padding=8
    
    shape = image.shape   
    image = tf.image.random_flip_left_right(image)
    image = tf.image.resize_with_crop_or_pad(image, shape[0]+padding, shape[1]+padding) 
    image = tf.image.random_crop(image, shape) 
    
    return image, target


def load_data(dataset='cifar10', batch_size=64):
    
    if dataset == 'cifar10': data = datasets.cifar10.load_data()
    elif dataset == 'cifar100': data = datasets.cifar100.load_data()
    else: raise("Unknown Dataset")
    
    (train_images, train_labels), (test_images, test_labels) = data

   
    train_images = np.asarray(train_images, np.float32)
    train_labels = np.asarray(train_labels, np.int32)

    test_images = np.asarray(test_images, np.float32)
    test_labels = np.asarray(test_labels, np.int32)
    
    num_train_data = train_images.shape[0]
   
    mean = channel_mean(train_images)
    std = channel_std(train_images)

    train_images = channel_standardization(train_images, mean, std)
    test_images = channel_standardization(test_images, mean, std)
    
    train_dataset = tf.data.Dataset.from_tensor_slices((train_images, train_labels))   
    train_dataset = train_dataset.shuffle(buffer_size=num_train_data, reshuffle_each_iteration=True)
    train_dataset = train_dataset.map(augment, num_parallel_calls=tf.data.experimental.AUTOTUNE)
    train_dataset = train_dataset.batch(batch_size)
    train_dataset = train_dataset.prefetch(4)
   
    
    test_dataset = tf.data.Dataset.from_tensor_slices((test_images, test_labels))
    test_dataset = test_dataset.batch(batch_size)
    test_dataset = test_dataset.prefetch(4)
    
    return train_dataset, test_dataset


if __name__ =='__main__':
    train_dataset, test_dataset = load_data('cifar100')
    
    num_data = 0
    for x, y in train_dataset:
        visualize(x[0,:],x[0,:])
        num_data += len(y)
