#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 22 17:05:52 2020

@author: 
"""

import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import tensorflow.keras as keras
from collections import defaultdict


def configure_gpu():
    
    gpus = tf.config.experimental.list_physical_devices('GPU')
    if gpus:
        try:
            # Currently, memory growth needs to be the same across GPUs
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
                logical_gpus = tf.config.experimental.list_logical_devices('GPU')
                print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            # Memory growth must be set before GPUs have been initialized
            print(e)
            

def featuremap_size(input_shape):
    input_shape = input_shape if isinstance(input_shape, (tuple, list)) else [input_shape]    
    channels = [shape[-1] for shape in input_shape]

    height = input_shape[0][1]
    width = input_shape[0][2]
    return channels, height, width




def safe_pow(x, p):
  x_ok = tf.greater(x, 0.)
  f = lambda x: tf.pow(x, p)
  
  safe_f = tf.zeros_like
  safe_x = tf.where(x_ok, x, tf.ones_like(x))
  
  return tf.where(x_ok, f(safe_x), safe_f(x))

def safe_norm(tensor, ord='euclidean', axis=None, keepdims=None, name=None):
    
    @tf.custom_gradient
    def norm(x):
        y = tf.norm(x, ord, axis, keepdims, name)
    
        def grad(dy):
            ex_dy = tf.expand_dims(dy, axis) if axis else dy
            ex_y = tf.expand_dims(y, axis)  if axis else y
            return ex_dy * (x/ (ex_y + 1e-19))
    
        return y, grad
    return norm(tensor)


def l2_norm(tensor, axis=None, keepdims=None, name=None):
    
    @tf.custom_gradient
    def norm(x):
        y = tf.norm(x, 2, axis, keepdims, name)
    
        def grad(dy):
            ex_dy = tf.expand_dims(dy, axis) if axis else dy
            ex_y = tf.expand_dims(y, axis)  if axis else y
            return ex_dy * (x/ (ex_y + 1e-19))
    
        return y, grad
    
    return norm(tensor)




class Plot_Dense_Conn(keras.callbacks.Callback):
    
    def __init__(self, **kwargs):
        super(Plot_Dense_Conn, self).__init__(**kwargs)
        
        self.history = []
        
    def _plot(self):

        
        batch_norms = self.model.get_batch_norm()
        rows = []        
        beta_list = defaultdict(list)
        
        max_length = 0
        
        for batch_norm in batch_norms:
        
                sizes, nonzeros, zeros, alpha, beta, reg = batch_norm.get_state()
                alpha = np.abs(np.squeeze(np.asarray(alpha)))

                alpha_max = np.max(alpha) 
                row = alpha / alpha_max if alpha_max > 0 else np.zeros_like(alpha)
                rows += [row] * 9 + [np.zeros_like(row)]

                max_length = len(row) if max_length < len(row) else max_length
                    
                beta_list[batch_norm.name].append(beta)
                
        figure = np.zeros(shape=(len(rows)+1, max_length))
        for i, row in enumerate(rows):
            figure[i+1,:row.size] = row
        
        #conns = np.abs(conns)
        #conns = conns / np.max(conns, axis=-1, keepdims=True)
        plt.figure(figsize=(15,15))
        plt.imshow(figure)
        plt.show()       
        
        #print('Betas: ', [(key, [np.array_str(np.squeeze(np.asarray(v))) for v in value]) for key, value in beta_list.items()] )

        return figure
    
    def on_train_begin(self, logs=None):
        figure = self._plot()
        self.history.append(figure)

    def on_epoch_end(self, epoch, logs=None):
        figure = self._plot()        
        self.history.append(figure)

class Plot_Group(keras.callbacks.Callback):
    
    def __init__(self, **kwargs):
        super(Plot_Group, self).__init__(**kwargs)
        
        self.history = []
        
    def _plot(self):

        
        batch_norms = self.model.get_batch_norm()
        rows = []
        
        max_length = 0
        
        mask = np.ones(shape=(10,)); mask[-1] = 0
        for batch_norm in batch_norms:
        
                sizes, nonzeros, zeros, alpha, beta, reg = batch_norm.get_state()

                nonzero_ratio  = [n/s for n, s in zip(nonzeros, sizes)]
                #nonzero_groups = np.squeeze(np.asarray(nonzero_groups))
                row = np.concatenate([np.zeros(shape=(1,))] + [r*mask for r in nonzero_ratio] , axis=-1)
                max_length = len(row) if max_length < len(row) else max_length
                rows += [row] * 9 + [np.zeros_like(row)]

                    
                
        figure = np.zeros(shape=(len(rows)+1, max_length))
        for i, row in enumerate(rows):
            #groups[i+1,:len(group)] = group
            figure[i+1,:row.size] = row


        plt.figure(figsize=(15,15))
        plt.imshow(figure)
        plt.show()       
        
        return figure
    
    def on_train_begin(self, logs=None):
        figure = self._plot()
        self.history.append(figure)

    def on_epoch_end(self, epoch, logs=None):
        figure = self._plot()        
        self.history.append(figure)
        
class Plot_Sparsity(keras.callbacks.Callback):
    
    def __init__(self, **kwargs):
        super(Plot_Sparsity, self).__init__(**kwargs)
            
        self._count_zero_channels = defaultdict(list)
        self._count_zero_groups = defaultdict(list)

        self._channel_sparsity_rates = defaultdict(list)
        self._group_sparsity_rates = defaultdict(list)

        self._sparsity_loss = defaultdict(list)
        
    def _print_alpha_beta(self):

        
        total_channels = 0
        total_zero_channels = 0
        total_groups = 0
        total_zero_groups = 0
        
        total_sparsity_loss = 0
        batch_norms = self.model.get_batch_norm()
        
        for batch_norm in batch_norms:
        
                sizes, nonzeros, zeros, alpha, beta, reg = batch_norm.get_state()
                
                channels = tf.reduce_sum(sizes).numpy()
                nonzero_channels = tf.reduce_sum(nonzeros).numpy()
                zero_channels = tf.reduce_sum(zeros).numpy()
                reg = reg.numpy()

                #print(alpha)
                #print(size, nonzeros, reg)
                
                total_channels += channels
                total_zero_channels += zero_channels
        
                groups = len(sizes)
                total_groups += groups
                zero_groups = tf.reduce_sum([0 if n > 0 else 1 for n in nonzeros]).numpy()
                total_zero_groups += zero_groups
                
                total_sparsity_loss += reg
                
                sparsity_rate = zero_channels/channels
                group_sparsity_rate = zero_groups / groups
                
                self._count_zero_channels[batch_norm.name].append(zero_channels)
                self._count_zero_groups[batch_norm.name].append(zero_groups)
                
                self._channel_sparsity_rates[batch_norm.name].append(sparsity_rate)
                self._group_sparsity_rates[batch_norm.name].append(group_sparsity_rate)

                self._sparsity_loss[batch_norm.name].append(reg)

                
        total_channel_sparsity_rate = total_zero_channels / total_channels if total_channels > 0 else 0
        total_group_sparsity_rate = total_zero_groups / total_groups if total_groups > 0 else 0

        
        self._count_zero_channels['total'].append(total_zero_channels)
        self._count_zero_groups['total'].append(total_zero_groups)

        self._channel_sparsity_rates['total'].append(total_channel_sparsity_rate)
        self._group_sparsity_rates['total'].append(total_group_sparsity_rate)

        self._sparsity_loss['total'].append(total_sparsity_loss)

        # plt.figure()
        # key = 'total'
        # value = self._sparsity_loss[key]
        # plt.plot(value, label=key)
            
        # plt.legend(loc='upper left')
        # plt.title('Sparsity Loss')            
        # plt.show()        
        
        #print('Sparsity Reg.: ', [(key, '%.4f'%value[-1]) for key, value in self._sparsity_loss.items()] )

        # plt.figure()
        # key = 'total'
        # value = self._count_zero_channels[key]
        # plt.plot(value, label=key)
            
        # plt.legend(loc='upper left')
        # plt.title('Zero Channels')            
        # plt.show()        
        
        #print('zero channels: ', [(key, value[-1]) for key, value in self._count_zero_channels.items()] )
        
        plt.figure()
        key = 'total'
        value = self._channel_sparsity_rates[key]
        plt.plot(value, label=key)
            
        plt.legend(loc='upper left')
        plt.title('Channel Sparity Rate')            
        plt.show()        
        #print('Channel Sparsity Rates.: ', [(key, '%.4f'%value[-1]) for key, value in self._channel_sparsity_rates.items()] )
        print('Channel Sparsity Rates, # zero channels, total # channels: ', total_channel_sparsity_rate, total_zero_channels, total_channels)
        
        # sparsity_per_layer = [value[-1] * 100 for key, value in self._channel_sparsity_rates.items()]
        # plt.figure()
        # plt.plot(sparsity_per_layer)
        # plt.title('Sparsity Rate per Layer')
        # plt.show()
        
        

        # plt.figure()
        # key = 'total'
        # value = self._count_zero_groups[key]
        # plt.plot(value, label=key)
            
        # plt.legend(loc='upper left')
        # plt.title('Zero Groups')            
        # plt.show()        
        
        #print('zero groups: ', [(key, value[-1]) for key, value in self._count_zero_groups.items()] )
        
        plt.figure()
        key = 'total'
        value = self._group_sparsity_rates[key]
        plt.plot(value, label=key)
            
        plt.legend(loc='upper left')
        plt.title('Group Sparity Rate')            
        plt.show()        
        #print('Group Sparsity Rates.: ', [(key, '%.4f'%value[-1]) for key, value in self._group_sparsity_rates.items()] )
        print('Group Sparsity rate, # zero groups, total # groups: ', total_group_sparsity_rate, total_zero_groups, total_groups)        
    
    def on_train_begin(self, logs=None):
        self._print_alpha_beta()

    def on_epoch_end(self, epoch, logs=None):
        self._print_alpha_beta()        
        
        # sparsity_rate = get_sparsity_rate(self.model)
        # print('Sparsity Rate:', sparsity_rate)
              
class Plot(keras.callbacks.Callback):
    

        
    def on_epoch_end(self, epoch, logs=None):
        
        history = self.model.history.history
        
        for name, values in history.items():
            
            plt.figure()
            plt.plot(values)
            plt.title(name)
            plt.show()
        
        print('Epoch:', epoch)
